/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

/*  Lempel-Ziv-Welch

    https://en.wikipedia.org/wiki/Lempel%E2%80%93Ziv%E2%80%93Welch
*/

#include "prelude.h"
#include "trie.h"
#include "lzw.h"

#define size 12
#define maxc (1 << size)
char buf[maxc+1];

// écrit le code c sur n chiffres en binaire
void output_bits(unsigned int n, unsigned int c, FILE *out) {
  if (n == 0) return;
  output_bits(n - 1, c >> 1, out);
  putc((c%2 == 1) ? '1' : '0', out);
}

void strsub(char *w, char *s, int ofs, int len) {
  memcpy(w, s+ofs, len);
  w[len] = '\0';
}

void encode(char *s, FILE *out) {
  trie *codes = trie_create();
  int next = 256;
  for (int c = 0; c < 256; c++) {
    buf[0] = c; buf[1] = '\0';
    trie_put(codes, buf, c);
  }
  while (*s != 0) {
    int p = trie_prefix(codes, s);
    strsub(buf, s, 0, p);
    int c = trie_get(codes, buf);
    output_bits(size, c, out);
    if (p < strlen(s) && next < maxc) {
      strsub(buf, s, 0, p+1);
      trie_put(codes, buf, next++);
    }
    s += p;
  }
}

int input(char *s) {
  assert(strlen(s) >= size);
  int v = 0;
  for (int i = 0; i < size; i++)
    v = 2*v + (s[i] == '1' ? 1 : 0);
  return v;
}

void decode(char *s, FILE *out) {
  assert(strlen(s) % size == 0);
  char *words[maxc];
  for (int c = 0; c < 256; c++) {
    words[c] = malloc(2);
    words[c][0] = c;
    words[c][1] = '\0';
  }
  int next = 256;
  int last = -1;
  while (*s != 0) {
    int c = input(s);
    assert(c <= next);
    if (last != -1 && next < maxc) {
      int len = strlen(words[last]);
      words[next] = malloc(len + 2);
      strcpy(words[next], words[last]);
      words[next][len] = words[(c == next) ? last : c][0];
      words[next][len + 1] = '\0';
      next++;
    }
    fputs(words[c], out); fflush(out);
    last = c;
    s += size;
  }
}
